/*
** Profile Emulation for Unix
** RCS Version $Revision: 1.16 $
** RCS Last Change Date: $Date: 2002/10/07 18:18:46 $
** Version 1.0.0.0
** Original Author: Michael I. Schwartz, mschwart@nyx.net
**
** {LICENSE}
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR WITH PERMISSION TO USE, COPY, MODIFY, AND
** DISTRIBUTE IT FOR ANY PURPOSE WITH THE FOLLOWING CONDITIONS:
** 1) IN SOURCE FORM, THIS HEADER MUST BE PRESERVED AND THESE CONDITIONS PROPOGATED
** 2) IN BINARY FORM, THE ORIGINAL AUTHOR MUST BE ACKNOWLEDGED IN DOCUMENTATION AND
**    CREDITS SCREEN (IF ANY)
** 3) FOR ANY COMMERCIAL SALE, THE AUTHOR MUST BE NOTIFIED OF THE USE OF THIS CODE
**    IN A PARTICULAR PRODUCT; THE ORIGINAL AUTHOR MAY CITE 
**    THE COMMERCIAL SELLER AS A BENEFICIARY OF THE CODE.
** 
** IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
** INDIRECT, SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF
** THE USE OF THIS SOFTWARE, ITS DOCUMENTATION,  OR ANY DERIVATIVES
** THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
** 
** THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO,  THE IMPLIED WARRANTIES OF
** MERCHANTABILITY,FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT. 
** THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND
** DISTRIBUTORS HAVE NO OBLIGATION  TO PROVIDE MAINTENANCE, SUPPORT,
** UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
** 
** {SYNOPSIS}
** This is profile code for Unix.
** The basic scheme is to have a directory (/usr/local/lib/TclProfiles) of .ini
** files, including a default file "win.ini"
** Each section of the file will be marked with square brackets
** Each entry in each section will comprise a word, and equal sign,
** and an arbitrary value.
** Blank lines and comment lines beginning with a semicolon (;) are permitted and ignored.
**
** There are four functions:
** const char *get_profile_dir();
** const char *set_profile_dir(const char *newdir);
** int get_profile_string ( const char *section,
** 			    const char *entry,
** 			    const char *defaultval,
** 			    char *outbuf,
** 			    size_t     outbufsiz,
** 			    const char *filename);
** int set_profile_string ( const char *section,
** 			    const char *entry,
** 			    const char *value,
** 			    const char *filename);
** get_profile_dir
**    Returns the current setting for the profile directory, by default
**    /usr/local/lib/TclProfiles
** set_profile_dir
**    Returns the current setting for the profile directory after
**    setting it to the newdir argument.
**    No attempt is made to determine if the setting is valid.
** get_profile_string
**    Returns the number of bytes needed to accept the result if outbufsiz is 0.
**    Otherwise, returns the number of bytes returned in the result.
**    Returns 0 upon failure.
**    The value in outbuf is one or more strings, each null terminated, followed by an
**    extra null character. The outbuf will not be written beyond outbuflen bytes.
**    If the filename is non-null, this is assumed to be a filename in the profile
**    directory overriding "win.ini", which is the default. No path check is made
**    for relative pathing.
**    If the section is non-null, specifies a section in which the rest of the
**    profile is examined.
**    If the section IS null, all the sections in the profile file are returned, and
**    all other values are ignored.
**    If the entry (and section) is non-null, the value for that entry is returned.
**    If the entry IS null, all entries within the selected section are returned.
** set_profile_string
**    Returns 0 upon failure, positive value upon success.
**    If the filename is non-null, this is assumed to be a filename in the profile
**    directory overriding "win.ini", which is the default. No path check is made
**    for relative pathing.
**    The section must be non-null, and specifies a section in which the rest of the
**    profile is examined.
**    The entry must be non-null, and specifies the entry within the section to
**    create or modify
**    The key must be non-null, and specifies the value to assign to the entry.
**    Note that the key can be blank, which creates an entry with a blank value.
**
**    These functions roughly match the functionality of Windows GetProfileString,
**    GetPrivateProfileString, SetProfileString, and SetPrivateProfileString
**    (close enough to port the profile Tcl package).
**
** {LOG}
** $Log: tkuxini.c $
** Revision 1.16  2002/10/07 18:18:46  Michael_Schwartz
** Changed extension name from profile to ini
**
** Revision 1.15  2002/10/07 17:38:09  Michael_Schwartz
** Updated version to 1.0.0.0
** Prepared to rename extension
**
** Revision 1.14  2001/07/26 03:24:10  Michael_Schwartz
** Added newlines to the new entries that are added. This apparently was
** lost in a previous build after CM/Test!
**
** Revision 1.12  1998/09/29  03:21:18  Michael_Schwartz
** *** empty log message ***
**
** Revision 1.11  1998/01/14  05:52:56  Michael_Schwartz
** *** empty log message ***
**
** Revision 1.10  1998/01/14  05:43:05  Michael_Schwartz
** Fixed problem with creating new .INI files (open fails).
**
** Revision 1.9  1997/10/19  22:21:02  Michael_Schwartz
** Add <sys/param.h> to get MAXPATHLEN
**
** Revision 1.8  1997/10/19  20:26:30  Michael_Schwartz
** Updated for first complete release.
** Handled blank lines in profile in a "sensible" manner
**
*/
#ifndef unix
#error "This code is only intended to be compiled for Unix systems"
#else
#include <sys/param.h>
#endif

#ifndef DEFAULT_PROFILE_DIR
#define DEFAULT_PROFILE_DIR "/usr/local/lib/TclProfiles"
#endif

#ifndef DEFAULT_PROFILE
#define DEFAULT_PROFILE "win.ini"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tkuxini.h"

#ifndef MAXDIRLEN
#define MAXDIRLEN MAXPATHLEN
#endif

/* If this code is to be used in a Tcl Extension, use the portable allocator */
#ifdef TCL_EXTENSION
#include <tcl.h>
#define calloc(a,b) Tcl_Alloc((a)*(b))
#define free(x)     Tcl_Free(x)
#endif

/* Internal helper functions */
static char *fgetline (char *buffer, int buflen, FILE *ifp);
static int fputline (const char *buffer, FILE *ofp);
 
static int delete_profile_section (const char *section, const char *filename);
static int delete_profile_entry   (const char *section, const char *entry, 
                                   const char *filename);

/* Local data */
static char default_dir[MAXDIRLEN]  = DEFAULT_PROFILE_DIR;
static const char default_profile[] = DEFAULT_PROFILE;
static int open_file_switch;

/****************************************************************
** While it might be easier to do this directly into Tcl by use
** of Tcl_AppendResult, this way there is a pretty close
** correspondence to the Windows routine GetProfileString, which
** might be useful.
****************************************************************/
int get_profile_string ( const char *section,
			 const char *entry,
			 const char *defaultval,
			 char *outbuf,
			 size_t     outbufsiz,
			 const char *filename)
{
  size_t outind = 0;
  FILE *pfile;
  char *buffer;
  char line[80];
  
  if (filename == 0)
    filename = default_profile;

  /* First, copy in the default string to handle the failure cases */
  strncpy (outbuf, defaultval, outbufsiz);
  outbuf[outbufsiz-1] = '\0';
  
  if ( (buffer = calloc(strlen(filename) + strlen(default_dir) + 2, sizeof(char)) ) == 0)
    /* Error! no memory! */
  {
    return 0;
  }
  
  sprintf(buffer,"%s/%s", default_dir, filename);
  pfile = fopen(buffer, "r");
  free (buffer);

  if ( pfile  == NULL )
    return 0;

  /* Now IF there is a section mentioned, look for it;
  ** otherwise, look for ALL sections
  */
  if (section)
  {
    int found_section = 0;

    while ( fgetline(line, sizeof(line), pfile) != 0 )
    {
      if ( line[0] == ';' )
        continue; /* Skip comment lines */
        
      if ( line[0] == '[')
      {
        char *cp = strtok(line,"]");
        if (cp == 0)
          continue;
        if (strcasecmp(cp+1, section) == 0)
        {
          int found_entry = 0;
            
          found_section = 1;
          /* Now IF there is an entry to look for, find it
          ** and return it; otherwise, return the list of entries
          */
          if (entry)
          {
            while ( fgetline(line, sizeof(line), pfile) != 0 &&
                    line[0] != '[')
            {
              char *cp;

              if ( line[0] == ';' )
                continue; /* Skip comment lines */
        
              if ( (cp = strtok(line, "=")) == 0)
                continue; /* Skip other lines without an = */
                
              if (strcasecmp(cp, entry) == 0)
              {
                found_entry = 1; /* Even if the entry is blank! */
                cp = strtok(NULL, "\n\r");
                if ( cp == 0 )  /* Blank entry */
                  cp = "";
                if (outbufsiz > 0)
                {
                  strncpy(outbuf, cp, outbufsiz);
                  outbuf[outbufsiz - 1] = '\0';
                }
                outind += strlen(cp) + 1; /* +1 is to cover the case of an empty entry */
                break;
              }
            }
            if (found_entry == 0) /* Nothing copied! */
            {
              if (outbufsiz > 0 )
              {
                strncpy(outbuf, defaultval, outbufsiz);
                outbuf[outbufsiz - 1] = '\0';
              }
              outind += strlen(defaultval);
            }
          }
          else
          {
            while ( fgetline(line, sizeof(line), pfile) != 0 &&
                    line[0] != '[')
            {
              char *tok;
              int  length;

              if ( line[0] == ';')
                continue; /* Skip comment lines */
              if ( (tok = strtok(line, "=")) == 0 )
                continue; /* Not a value line (a blank line?) */

              length = strlen(tok);
              if (outbufsiz - outind > length + 1)
              {
                strcpy(outbuf+outind, tok);
                outind += length + 1;
              }
              else if (outbufsiz == 0)
                outind += length + 1;
            }
            if (outbufsiz > 0)
              outbuf[outind] = '\0'; /* Extra nul */
            outind++;
          }
          break;
        }
      }
    }
    if ( found_section == 0 ) /* No such section--use the default! */
    {
      if (outbufsiz > 0 )
      {
        strncpy(outbuf, defaultval, outbufsiz);
        outbuf[outbufsiz - 1] = '\0';
      }
      outind += strlen(defaultval);
    }
  }
  else
  {
    while ( fgetline(line, sizeof(line), pfile) != 0 )
    {
      if ( line[0] == '[')
      {
        char *tok;
        int  length;

        if ( (tok = strtok(line,"]")) == 0 )
          continue;
        tok++;   /* Do not include the [ in the section name */

        length = strlen(tok);
        if (outbufsiz - outind > length + 1)
        {
          strcpy(outbuf+outind, tok);
          outind += length + 1;
        }
        if (outbufsiz == 0)
          outind += length + 1;
      }
    }
    if (outbufsiz > 0)
      outbuf[outind] = '\0'; /* Extra nul */
    outind++;
  }
  fclose(pfile);
    
  return outind;
}

/****************************************************************
** The processing is as follows:
** If filename is non-0, the given file is used (relative to
** the profile directory) rather than the default file (win.ini)
** If the section is non-zero, it represents the section affected.
** If the section is zero, the request is illegal.
** If the entry is non-zero, it represents the entry affected.
** If the entry is zero, the entire section is deleted.
** If the value is non-zero, it represents the new value to
** assign to the entry.
** If the value is zero, the entry is deleted.
****************************************************************/
int set_profile_string ( const char *section,
			 const char *entry,
			 const char *value,
			 const char *filename)
{
  char *ibuffer, *obuffer;
  char line[256], temp[256];
  FILE *ipfile, *opfile;
  int offset;
  int status;
  int foundsection=0;
  int foundkey = 0;
  
  if ( section == 0 )
    return 0; /* Error! invalid request */

  if (filename == 0)
    filename = default_profile;

  if ( entry == 0 )
    return delete_profile_section (section, filename);

  if ( value == 0 )
    return delete_profile_entry   (section, entry, filename);
 
  if ( (ibuffer = calloc(strlen(filename) + strlen(default_dir) + 2, sizeof(char)) ) == 0)
    /* Error! no memory! */
    return 0;

  if ( (obuffer = calloc(strlen(filename) + strlen(default_dir) + 2, sizeof(char)) ) == 0)
    /* Error! no memory! */
  {
    free(ibuffer);
    return 0;
  }

  sprintf(ibuffer,"%s/%s", default_dir, filename);

  sprintf(obuffer,"%s/%s", default_dir, filename);
  offset = strlen(obuffer) - 1; /* Subtract 1 to get to last in string */
  if ( obuffer[offset] != 'n' )
    obuffer[offset] = 'n'; /* .inn instead of .ini */
  else
    obuffer[offset] = 'o'; /* .xxo instead of .xxn (not a standard .ini file) */

  
  ipfile = fopen(ibuffer, "r"); /* It's OK if this fails--just means a new .ini file */

  if ( (opfile = fopen(obuffer, "w")) == NULL )
  {
    free(ibuffer);
    free(obuffer);
    return 0;
  }
  open_file_switch = -1;  /* Note that the file has been opened for write */

  /* From here, find the section */
  while ( ipfile && fgetline(line, sizeof(line), ipfile) != 0 )
  {
    if ( foundsection == 0 )
    {
      fputline (line, opfile);

      if ( line[0] == '[')
      {
        char *cp;

        cp = strtok(line,"]");
        if (cp == 0)
          continue;
        if (strcasecmp(cp+1, section) == 0)
        {
          /* If we've found the section, look for the key */
          foundsection = 1;
        }
      }
    }
    else if ( foundkey == 0 )
    {
      if ( line[0] == '[' )
      {
        sprintf(temp, "%s=%s\n", entry, value);
        fputline(temp, opfile);
        fputline(line, opfile);
        fputline("\n",  opfile);
        foundkey = 1;
      }
      else
      {
        char *cp;
        
        cp = strtok(line, "=");
        if ( strcasecmp (cp, entry) == 0 )
        {
          fputline(cp,  opfile );
          fputline("=", opfile );
          fputline(value, opfile);
          fputline("\n",  opfile);
          foundkey = 1;
        }
        else
        {
          fputline(cp, opfile);
          if ( (cp = strtok(NULL, "")) != NULL )
          {
            fputline("=", opfile);
            fputline(cp,   opfile);
            fputline("\n",  opfile);
          }
        }
      }
    }
    else
    {
      fputline (line, opfile);
      fputline("\n",  opfile);
    }
  }
  
  /* If end-of file, just create the section here, and we're done */
  if ( foundsection == 0 )
  {
    sprintf(line, "[%s]\n", section);
    fputline(line, opfile);
  }
  
  if ( foundkey == 0 )
  {
    sprintf(line, "%s=%s\n", entry, value);
    fputline(line, opfile);
  }
  
  if (ipfile)
    fclose(ipfile);
  fclose(opfile);

  remove ("tmp.tmp");
  rename (ibuffer, "tmp.tmp");
  rename (obuffer, ibuffer);
  remove ("tmp.tmp");

  free(ibuffer);
  free(obuffer);
  return 1; /* success? */
}

/****************************************************************
** This function is intended for rare invocation, so thread
** safety is not provided.
****************************************************************/
const char *get_profile_dir (void)
{
  return default_dir;
}

/****************************************************************
** This function is intended for rare invocation, so thread
** safety is not provided.
****************************************************************/
const char *set_profile_dir (const char *profile_dir)
{
  strncpy(default_dir, profile_dir, sizeof(default_dir));
  default_dir[sizeof(default_dir) - 1] = '\0';
  return default_dir;
}
			 

/****************************************************************
** Delete_profile_section
****************************************************************/
static int delete_profile_section (const char *section, const char *filename)
{
  char *ibuffer, *obuffer;
  char line[256], temp[256];
  FILE *ipfile, *opfile;
  int foundsection=0;
  size_t offset;
  
  if ( (ibuffer = calloc(strlen(filename) + strlen(default_dir) + 2, sizeof(char)) ) == 0)
    /* Error! no memory! */
    return 0;

  if ( (obuffer = calloc(strlen(filename) + strlen(default_dir) + 2, sizeof(char)) ) == 0)
    /* Error! no memory! */
  {
    free(ibuffer);
    return 0;
  }

  sprintf(ibuffer,"%s/%s", default_dir, filename);

  sprintf(obuffer,"%s/%s", default_dir, filename);
  offset = strlen(obuffer) - 1; /* Subtract 1 to get to last in string */
  if ( obuffer[offset] != 'n' )
    obuffer[offset] = 'n'; /* .inn instead of .ini */
  else
    obuffer[offset] = 'o';
  
  if ( (ipfile = fopen(ibuffer, "r")) == NULL )
  {
    free(ibuffer);
    free(obuffer);
    return 1;     /* If no .ini file exists, deletion is completed successfully... */
  }

  if ( (opfile = fopen(obuffer, "w")) == NULL )
  {
    free(ibuffer);
    free(obuffer);
    return 0;
  }
  open_file_switch = -1;  /* Note that the file has been opened for write */

  /* From here, find the section */
  while ( fgetline(line, sizeof(line), ipfile) != 0 )
  {
    if ( foundsection == 0 )
    {
      strcpy (temp, line);

      if ( line[0] == '[')
      {
        char *cp;

        cp = strtok(line,"]");
        if (cp == 0)
        {
          /* This is impossible, but output it anyway. */
          fputline(temp, opfile);
          continue;
        }
        
        if (strcasecmp(cp+1, section) != 0)
        {
          /* It's a section, but not the one we're looking for */
          fputline (temp, opfile);
        }
        else
        {
          foundsection = 1; /* Found our section */

          /* Don't output this section header */
          
          /* Cycle until the section is complete. */
          while ( fgetline(line, sizeof(line), ipfile) != 0 )
          {
            if ( line[0] == '[' )
            {
              fputline(line, opfile);
              break;
            }
            /* Otherwise, ignore the line (no output). */
          }
        } /* End of our section case */
      } /* End of it's some section case */
      else
        fputline(temp, opfile);
    } /* End of we haven't found our section case */
    else
    {
      fputline(line, opfile);
    }
  }

  fclose(ipfile);
  fclose(opfile);

  remove ("tmp.tmp");
  rename (ibuffer, "tmp.tmp");
  rename (obuffer, ibuffer);
  remove ("tmp.tmp");

  free(ibuffer);
  free(obuffer);

  return 1;
}

/****************************************************************
** Delete_profile_entry
****************************************************************/
static int delete_profile_entry   (const char *section, const char *entry,
                                   const char *filename)
{
  char *ibuffer, *obuffer;
  char line[256], temp[256];
  FILE *ipfile, *opfile;
  int offset;
  int status;
  int foundsection=0;
  int foundkey = 0;
  
  if ( (ibuffer = calloc(strlen(filename) + strlen(default_dir) + 2, sizeof(char)) ) == 0)
    /* Error! no memory! */
    return 0;

  if ( (obuffer = calloc(strlen(filename) + strlen(default_dir) + 2, sizeof(char)) ) == 0)
    /* Error! no memory! */
  {
    free(ibuffer);
    return 0;
  }

  sprintf(ibuffer,"%s/%s", default_dir, filename);

  sprintf(obuffer,"%s/%s", default_dir, filename);
  offset = strlen(obuffer) - 1; /* Subtract 1 to get to last in string */
  if ( obuffer[offset] != 'n' )
    obuffer[offset] = 'n'; /* .inn instead of .ini */
  else
    obuffer[offset] = 'o';
  
  if ( (ipfile = fopen(ibuffer, "r")) == NULL )
  {
    free(ibuffer);
    free(obuffer);
    return 1; /* If there is no profile, deletion is successful */
  }

  if ( (opfile = fopen(obuffer, "w")) == NULL )
  {
    free(ibuffer);
    free(obuffer);
    return 0;
  }
  open_file_switch = -1; /* Note that the file has been opened for write */

  /* From here, find the section */
  while ( fgetline(line, sizeof(line), ipfile) != 0 )
  {
    if ( foundsection == 0 )
    {
      fputline (line, opfile);

      if ( line[0] == '[')
      {
        char *cp;

        cp = strtok(line,"]");
        if (cp == 0)
          continue;
        if (strcasecmp(cp+1, section) == 0)
        {
          /* If we've found the section, look for the key */
          foundsection = 1;
        }
      }
    }
    else if ( foundkey == 0 )
    {
      if ( line[0] == '[' )
      {
        fputline(line, opfile);
        foundkey = 1;   /* The key doesn't exist--thus, it is deleted */
      }
      else
      {
        char *cp;
        strcpy (temp, line);
        
        cp = strtok(line, "=");
        if ( strcasecmp (cp, entry) == 0 )
          foundkey = 1;
        else
          fputline(temp, opfile);
      }
    }
    else
    {
      fputline (line, opfile);
    }
  }

  fclose(ipfile);
  fclose(opfile);

  remove ("tmp.tmp");
  rename (ibuffer, "tmp.tmp");
  rename (obuffer, ibuffer);
  remove ("tmp.tmp");

  free(ibuffer);
  free(obuffer);

  return 1;
}

/****************************************************************
** fgetline -- like fgets, but ignores blank lines
****************************************************************/
static char *fgetline (char *buffer, int buflen, FILE *ifp)
{
  char *retval;
  while ( ( retval = fgets(buffer, buflen, ifp ) ) != 0 )
  {
    if ( buffer[0] != '\n' && buffer[0] != '\r' )
      return retval;
  }
  return retval;
}

/****************************************************************
** fputline -- like fputs, but adds a blank line before each section
** after the first one
****************************************************************/
static int fputline (const char *buffer, FILE *ofp)
{
  char *retval;

  if (buffer[0] == '[' && open_file_switch == 1 )
    fputs("\n", ofp);

  open_file_switch = 1;
  return fputs (buffer, ofp);
}
